"""
Module for batch-handling windows registry.
WARNING: Not Unicode-safe
"""

#IMPORTS
#For exception info
from sys import exc_info
#Windows Registry Access
import _winreg
#For getenv("temp"), unlink(), popen()
import os


class RegError(Exception):
    pass

class RegExpandStr(unicode):
    def __init__(self, src):
        unicode.__init__(self, src)

class RegBinaryStr(str):
    def __init__(self, src):
        str.__init__(self, src)

class RegKey():
    def __init__(self, name = ''):
        self.name = name
        self.values = {}
        self.subkeys = []

    def __repr__(self):
        return self.__str__()

    def __str__(self, indent = 0):
        def _str_list(lst):
            return "\n".join([n.__str__(indent) for n in lst])
        indent += 1
        indentStr = "    " * (indent-1)
        formatStr = indentStr+"<RegKey>\n"+indentStr+"Name: %s\n"+indentStr+"Values: %s\n"+indentStr+"Subkeys:\n%s"
        str = formatStr % (self.name, self.values, _str_list(self.subkeys))

        indent -= 1
        return str

    def findSubkey(self, name):
        i = 0

        for subkey in self.subkeys:
            if subkey.name.lower() == name.lower():
                return i
            i += 1

        return None

    def getSubkey(self, name):
        i = 0

        for subkey in self.subkeys:
            if subkey.name.lower() == name.lower():
                return subkey
            i += 1

        raise IndexError, "No such subkey."

    def deleteSubkey(self, name):
        i = 0

        for subkey in self.subkeys:
            if subkey.name.lower() == name.lower():
                self.subkeys.remove(subkey)
                return
            i += 1

        raise IndexError, "No such subkey."

    def merge(self, otherkey):
        #Update values
        self.values.update(otherkey.values)

        #Update subkeys
        for foreignSubkey in otherkey.subkeys:
            subkeyIndex = self.findSubkey(foreignSubkey.name)
            if subkeyIndex is None:
                self.subkeys.append(foreignSubkey)
            else:
                self.subkeys[subkeyIndex].merge(foreignSubkey)


    def toScript(self, path, addHeader = True):
        #Returns a windows .reg script that can be run to add
        #the key (including values and subkeys).
        
        lines = []

        #Header
        if addHeader:
            lines.append("Windows Registry Editor Version 5.00\n")
        else:
            lines.append("")

        #Main key
        lines.append("[%s\\%s]" % (path, self.name))

        #Own values
        for name, value in self.values.items():
            name = name and ('"'+name+'"') or "@"
            lines.append('%s=%s' % (name, RegEncode(value)))

        #Subkeys
        for subkey in self.subkeys:
             lines.append(subkey.toScript(path + "\\" + self.name, False))

        return "\n".join(lines)
        

def RegType(value):
    """
    Returns an int constant specifying the type of the inputted value.
    """
    if type(value) == str or type(value) == unicode:
        return _winreg.REG_SZ
    elif type(value) == int:
        return _winreg.REG_DWORD
    elif type(value) == list:
        return _winreg.REG_MULTI_SZ
    elif type(value) == RegBinaryStr:
        return _winreg.REG_BINARY
    elif type(value) == RegExpandStr:
        return _winreg.REG_EXPAND_SZ
    elif type(value) == type(None):
        return _winreg.REG_NONE
    else:
        raise RegError, "Unknown Data Type. (%s -> %s)" % (value, type(value))

def RegEncode(value):
    """
    Encodes the raw value to a string that can be used
    in a registry script.
    """
    if RegType(value) == _winreg.REG_SZ:
        value = value.replace('\\','\\\\')
        value = value.replace('"', '\\"')
        value = value.replace('\n','\\n')
        value = value.replace('\t','\\t')
        return '"%s"' % value
    elif RegType(value) == _winreg.REG_DWORD:
        value = "dword:%08x" % value
        return value
    elif RegType(value) == _winreg.REG_MULTI_SZ:
        valueBytes = []
        for item in value: #list
            valueBytes += ["%x,00" % ord(c) for c in item]
            valueBytes += ["00,00"]
        return  'hex(%d):%s,00,00' % (_winreg.REG_MULTI_SZ, ','.join(valueBytes))
    elif RegType(value) == _winreg.REG_NONE:
        return  'hex(%d):' % (_winreg.REG_NONE)
    else:
        valueBytes = ["%x" % ord(c) for c in value]
        return  'hex(%d):%s' % (RegType(value), ','.join(valueBytes))

def _RegLockKeySingle(path):
    path = path.replace("HKEY_CLASSES_ROOT", r"HKEY_LOCAL_MACHINE\Software\Classes")

    script = "%s [2 5 8 13 18]" % path
    scriptFile = "%s\\python.regbatch.lock.script" % os.getenv("temp", "C:/")
    try:
        hScriptFile = open(scriptFile, "w")
        hScriptFile.write(script)
    finally:
        try:
            hScriptFile.close()
        except:
            pass

    cmd = os.popen('regini "%s"' % scriptFile)
    cmd.close()
    
    try:
        os.unlink(scriptFile)
    except:
        pass

def _RegUnlockKeySingle(path):
    path = path.replace("HKEY_CLASSES_ROOT", r"HKEY_LOCAL_MACHINE\Software\Classes")

    script = "%s [1 5 13 18]" % path
    scriptFile = "%s\\python.regbatch.lock.script" % os.getenv("temp", "C:/")
    try:
        hScriptFile = open(scriptFile, "w")
        hScriptFile.write(script)
    finally:
        try:
            hScriptFile.close()
        except:
            pass

    cmd = os.popen('regini "%s"' % scriptFile)
    cmd.close()
    
    try:
        os.unlink(scriptFile)
    except:
        pass

def RegReadKey(path, parent = 0):
    """
    Recursively reads a registry key and its subkeys and
    returns a RegKey structure of all values.
    """
    #Get the parent node if not specified
    if not parent:
        (root, _junk, path) = path.partition("\\")
        try:
            parent = getattr(_winreg, root.upper())
        except WindowsError:
            raise RegError, "Invalid root node."

    #Create a RegKey object
    keyData = RegKey(path.split("\\")[-1])

    #Open target key
    try:
        hkey = _winreg.OpenKey(parent, path)
    except WindowsError:
        raise RegError, exc_info()[1]

    #Get values
    try:
        i = 0
        while True:
            (value_name, value_data, value_type) = _winreg.EnumValue(hkey, i)
            #Convert ambigous values to proper types
            if value_type == _winreg.REG_EXPAND_SZ:
                value_data = RegExpandStr(value_data)
            elif value_type == _winreg.REG_BINARY:
                value_data = RegBinaryStr(value_data)

            keyData.values[value_name] = value_data
            i += 1
    except WindowsError:
        pass

    #Recursively get subkeys
    try:
        i = 0
        while True:
            subkey = _winreg.EnumKey(hkey, i)
            keyData.subkeys.append(RegReadKey(subkey, hkey))
            i += 1
    except WindowsError:
        pass

    #Clean up!
    try:
            hkey.close()
    except:
            pass

    #Return the filled RegKey structure
    return keyData



def RegDeleteKey(path, parent = 0):
    """
    Recursively deletes a registry key and its subkeys.
    """
    #Get the parent node if not specified
    if not parent:
        (root, _junk, path) = path.partition("\\")
        try:
            parent = getattr(_winreg, root.upper())
        except:
            raise RegError, "Invalid root node."

    #Open the key in question so we can get its subkeys and check for access
    try:
        hkey = _winreg.OpenKey(parent, path)
    except WindowsError:
        raise RegError, exc_info()[1]

    #Recursively remove subkeys then delete the original
    try:
        while True:
            subkey = _winreg.EnumKey(hkey, 0)
            RegDeleteKey(subkey, hkey)
    except WindowsError:
        _winreg.DeleteKey(hkey,'')


def RegWriteKey(path, keyData, merge = True, parent = 0):
    """
    Recursively writes values to a registry key and its
    subkeys given a RegKey structure of all values. If the merge
    parameter is 0, the original key will be deleted before adding
    the new values.
    """
    #Get the parent node if not specified
    if not parent:
        path = "%s\\%s" % (path, keyData.name)
        (root, _junk, path) = path.partition("\\")
        try:
            parent = getattr(_winreg, root.upper())
        except:
            raise RegError, "Invalid root node."

    #Delete old key if not merging
    if not merge:
        try:
            RegDeleteKey(path, parent)
        except:
            pass

    try:
        #Create/Open Key
        hkey = _winreg.CreateKey(parent, path)

        #Set values
        for (value_name, value_data) in keyData.values.items():
            _winreg.SetValueEx(hkey, value_name, 0, RegType(value_data), value_data)
    except:
        raise RegError, "Access Denied"

    #Recursively fill subkeys
    for subkey in keyData.subkeys:
        RegWriteKey(subkey.name, subkey, True, hkey)

    #Clean up!
    try:
            hkey.close()
    except:
            pass

def RegLockKey(path):
    def _RegLockKey2(path, key):
        for subkey in key.subkeys:
            _RegLockKey2(path + "\\" + key.name, subkey)
            
        _RegLockKeySingle(path + "\\" + key.name)

        
    keyToLock = RegReadKey(path)
    _RegLockKey2(path.rpartition("\\")[0], keyToLock)

def RegUnlockKey(path):
    def _RegUnlockKey2(path, key):
        for subkey in key.subkeys:
            _RegUnlockKey2(path + "\\" + key.name, subkey)
            
        _RegUnlockKeySingle(path + "\\" + key.name)

        
    keyToLock = RegReadKey(path)
    _RegUnlockKey2(path.rpartition("\\")[0], keyToLock)
    

def RegRunScript(filename, silent = True):
    cmd = os.popen('regedit ' + (silent and '/s ' or '') + '"%s"' % (filename))
    cmd.close()

def EMERGENCY_EXE_FIX():
    hkey = _winreg.CreateKey(_winreg.HKEY_CLASSES_ROOT,'.exe\\shell\\open\\command')
    _winreg.SetValueEx(hkey,'',0,_winreg.REG_SZ,'"%1" %*')

#MAIN
if __name__ == '__main__':
    print RegReadKey(r"HKEY_CLASSES_ROOT\.aif").toScript('MY_ROOT')
    #RegUnlockKey("HKEY_CLASSES_ROOT\\.mp2")

