#sfa2ext - SystemFileAssociations -> Extension
#
#Moves an action defined in SytemFileAssociations for a particular
#PerceivedType to the classes of all the extensions of that type.

#Algorithm
#   Get action key from sfa/perceivedtype
#   Find all exts of perceived type
#   For each ext:
#       Check ext for action
#       Check class for action
#       if OVERWRITE
#           Delete action from class, ext if exists
#           Add action to class
#       else
#           If neither have action
#               Add action to class
#       if LOCK_WHEN_DONE
#           Lock ext, class



#Path, folder actions
import os
import sys
#High level registry access
from regbatch import *
#Low level registry access
import _winreg

HKCR = "HKEY_CLASSES_ROOT"

PERCEIVED_TYPE  = 'generic'
ACTION          = '_max.edit.bin'
OVERWRITE       = True
LOCK_WHEN_DONE  = True


def getRelatedKeys(ext):
    #Get extension key
    try:
        extKey = RegReadKey("%s\\.%s" % (HKCR, ext))
    except RegError:
        extKey = None
        #print "%s: Not found." % (ext)

    #Get class key
    try:
        classKey = RegReadKey("%s\\%s" % (HKCR, extKey.values['']))
    except RegError:
        classKey = None
        #print "%s: Class not found." % (ext)
    except AttributeError:
        classKey = None
        #print "%s: Unclassed." % (ext)
    except KeyError:
        classKey = None

    return (extKey, classKey)


def createExtKey(ext, name, handler, relatedKeys):
    oldExtKey = relatedKeys[0]
    
    #Create basic extension key
    extKey = RegKey(".%s" % ext)

    #Add class name
    extKey.values[''] = "_max.%s" % ext
    
    #Add PerceivedType
    extKey.values['PerceivedType'] = PERCEIVED_TYPE
        
            
    #Get Content-Type if one exists or create our own
    if oldExtKey:
        if oldExtKey.values.has_key('Content Type'):
            extKey.values['Content Type'] = oldExtKey.values['Content Type'] 
    if not extKey.values.has_key('Content Type'):
        extKey.values['Content Type'] = '%s/x-max-%s' % (PERCEIVED_TYPE, ext)
        

    #Get PersistentHandler if any
    if oldExtKey:
        index = oldExtKey.findSubkey('PersistentHandler')
        if index is not None:
            extKey.subkeys.append(oldExtKey.subkeys[index])
        
    return extKey

def moveIcons():
    '''
    Moves icons from the extension key to its class
    '''
    exts = []
    
    #Find all exts of perceived type
    i = 1 #Start from the second key, ignoring "*"
    while True:
        ext = _winreg.EnumKey(_winreg.HKEY_CLASSES_ROOT, i)
        if ext[0] != ".":
            break
        
        hkey = _winreg.OpenKey(_winreg.HKEY_CLASSES_ROOT, ext)
        try:
            curType = _winreg.QueryValueEx(hkey, "PerceivedType")[0]
            if curType == PERCEIVED_TYPE:
                exts.append(ext[1:])
        except:
            pass
        hkey.Close()
        i += 1

    #For each ext:
    for ext in exts:
        (extKey, classKey) = getRelatedKeys(ext)

        #if ext has an icon but class has none
        if (extKey.findSubkey('DefaultIcon') is not None) and (classKey.findSubkey('DefaultIcon') is None):
            classKey.subkeys.append(extKey.getSubkey('DefaultIcon'))
            extKey.deleteSubkey('DefaultIcon')

            try:
                RegWriteKey(HKCR, extKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + extKey.name)
                RegWriteKey(HKCR, extKey, False)
                
            try:
                RegWriteKey(HKCR, classKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + classKey.name)
                RegWriteKey(HKCR, classKey, False)

        if LOCK_WHEN_DONE:
            RegLockKey(HKCR + "\\" + extKey.name)
            RegLockKey(HKCR + "\\" + classKey.name)

def main():
    exts = []
    
    #Get action key from sfa/perceivedtype
    try:
        actionKeyPath = "%s\\SystemFileAssociations\\%s\\shell\\%s" % (HKCR, PERCEIVED_TYPE, ACTION)
        actionKey = RegReadKey(actionKeyPath)
    except RegError:
        raise RegError, "Type or action not found."

    #Find all exts of perceived type
    i = 1 #Start from the second key, ignoring "*"
    while True:
        ext = _winreg.EnumKey(_winreg.HKEY_CLASSES_ROOT, i)
        if ext[0] != ".":
            break
        
        hkey = _winreg.OpenKey(_winreg.HKEY_CLASSES_ROOT, ext)
        try:
            curType = _winreg.QueryValueEx(hkey, "PerceivedType")[0]
            if curType == PERCEIVED_TYPE:
                exts.append(ext[1:])
        except:
            pass
        hkey.Close()
        i += 1

    #For each ext:
    for ext in exts:
        (extKey, classKey) = getRelatedKeys(ext)
        
        #Check ext for action
        try:
            extHasAction = (extKey.getSubkey("shell").findSubkey(ACTION) is not None)
        except IndexError:
            extHasAction = False
            
        #Check class for action
        try:
            classHasAction = (classKey.getSubkey("shell").findSubkey(ACTION) is not None)
        except IndexError:
            classHasAction = False
            
        if OVERWRITE:
            #Delete action from class, ext if exists
            if extHasAction:
                extShellIndex = extKey.findSubkey("shell")
                extKey.subkeys[extShellIndex].deleteSubkey(ACTION)
            if classHasAction:
                classShellIndex = classKey.findSubkey("shell")
                classKey.subkeys[classShellIndex].deleteSubkey(ACTION)
                
            #Add action to class
            classShellIndex = classKey.findSubkey("shell")
            if classShellIndex is None:
                classKey.subkeys.append(RegKey("shell"))
                classShellIndex = classKey.findSubkey("shell")

            classKey.subkeys[classShellIndex].subkeys.append(actionKey)

            try:
                RegWriteKey(HKCR, extKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + extKey.name)
                RegWriteKey(HKCR, extKey, False)
                
            try:
                RegWriteKey(HKCR, classKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + classKey.name)
                RegWriteKey(HKCR, classKey, False)
                

        else:
            if (not extHasAction) and (not classHasAction):
                #Add action to class
                classShellIndex = classKey.findSubkey("shell")
                if classShellIndex is None:
                    classKey.subkeys.append(RegKey("shell"))
                    classShellIndex = classKey.findSubkey("shell")

                classKey.subkeys[classShellIndex].subkeys.append(actionKey)

            try:
                RegWriteKey(HKCR, extKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + extKey.name)
                RegWriteKey(HKCR, extKey, False)
                
            try:
                RegWriteKey(HKCR, classKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + classKey.name)
                RegWriteKey(HKCR, classKey, False)

        if LOCK_WHEN_DONE:
            RegLockKey(HKCR + "\\" + extKey.name)
            RegLockKey(HKCR + "\\" + classKey.name)

def main2():
    '''
    Moves icons from the extension key to its class
    '''
    exts = []
    
    #Find all exts of perceived type
    i = 1 #Start from the second key, ignoring "*"
    while True:
        ext = _winreg.EnumKey(_winreg.HKEY_CLASSES_ROOT, i)
        if ext[0] != ".":
            break
        
        hkey = _winreg.OpenKey(_winreg.HKEY_CLASSES_ROOT, ext)
        try:
            curType = _winreg.QueryValueEx(hkey, "PerceivedType")[0]
            if curType != PERCEIVED_TYPE:
                exts.append(ext[1:])
        except:
            exts.append(ext[1:])
        hkey.Close()
        i += 1

    #For each ext:
    for ext in exts:
        extKey = RegReadKey("%s\\.%s" % (HKCR, ext))

        if not extKey.values.has_key('PerceivedType'):
            extKey.values['PerceivedType'] = 'generic'
            
            try:
                RegWriteKey(HKCR, extKey, False)
            except RegError:
                RegUnlockKey(HKCR + "\\" + extKey.name)
                RegWriteKey(HKCR, extKey, False)

        if LOCK_WHEN_DONE:
            RegLockKey(HKCR + "\\" + extKey.name)

        
if __name__ == '__main__':
    os.chdir(sys.path[0])
    try:
        main()
    except:
        raise
    finally:
        try:
            emergencyTester = RegReadKey('HKEY_CLASSES_ROOT\\.exe')
            del emergencyTester
        except:
            EMERGENCY_EXE_FIX()



            
    


                    